

from mocat.src.scenarios.twodim.vectorise import TwoDimScenario

from mocat.src.scenarios.twodim.toy_scenarios import Gaussian
from mocat.src.scenarios.twodim.toy_scenarios import Banana
from mocat.src.scenarios.twodim.toy_scenarios import NealFunnel
from mocat.src.scenarios.twodim.toy_scenarios import GaussianMixture
from mocat.src.scenarios.twodim.toy_scenarios import DoubleWell
from mocat.src.scenarios.twodim.toy_scenarios import Rastrigin

from mocat.src.scenarios.twodim.visualise import visualise
from mocat.src.scenarios.twodim.visualise import RunVis
from mocat.src.scenarios.twodim.visualise import RunVisUtils



