
from mocat import twodim

from mocat.src.scenarios.toy_scenarios import Gaussian
from mocat.src.scenarios.toy_scenarios import GaussianMixture
from mocat.src.scenarios.toy_scenarios import DoubleWell
from mocat.src.scenarios.toy_scenarios import Rastrigin



